"""\
Single player system support module.

@author: Aaron Mavrinac (ezod)
@organization: Thousand Parsec
@license: GPL-2
"""

# Python imports
import os
import sys
import time
import socket
import urllib

# find an elementtree implementation
ET = None
errors = []
try:
    import elementtree.ElementTree as ET
except ImportError, e:
    errors.append(e)
try:
    import cElementTree as ET
except ImportError, e:
    errors.append(e)
try:
    import lxml.etree as ET
except ImportError, e:
    errors.append(e)
try:
    import xml.etree.ElementTree as ET
except ImportError, e:
    errors.append(e)
if ET is None:
    raise ImportError(str(errors))

# import from local 2.6 subprocess module
from subprocess import Popen

# local imports
import version


class _Server(dict):
	"""\
	Dictionary subclass for server descriptions.
	"""
	def __init__(self):
		for k in ['longname', 'version', 'description', 'commandstring', 'cwd']:
			self[k] = ''
		self['forced'] = []
		self['parameter'] = {}
		self['ruleset'] = {}
	
class _AIClient(dict):
	"""\
	Dictionary subclass for AI client descriptions.
	"""
	def __init__(self):
		for k in ['longname', 'version', 'description', 'commandstring', 'cwd']:
			self[k] = ''
		self['rules'] = []
		self['forced'] = []
		self['parameter'] = {}

class _Ruleset(dict):
	"""\
	Dictionary subclass for ruleset descriptions.
	"""
	def __init__(self):
		for k in ['longname', 'version', 'description']:
			self[k] = ''
		self['forced'] = []
		self['parameter'] = {}

class _Parameter(dict):
	"""\
	Dictionary subclass for parameter descriptions.
	"""
	def __init__(self):
		for el in ['type', 'longname', 'description', 'default', 'commandstring']:
			self[el] = ''


class LocalList(dict):
	"""\
	Local list of servers, rulesets, and AI clients.
	"""

	def __init__(self):
		"""\
		Constructor.
		"""
		self['server'] = {}
		self['aiclient'] = {}

		# look for installed single player XML files
		ins_sharepath = []
		if sys.platform == 'win32':
			try:
				import _winreg
				tpsp = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, "Software\\Thousand Parsec\\SinglePlayer")
				i = 0
				while True:
					name, value, t = _winreg.EnumValue(tpsp, i)
					ins_sharepath.append(value)
					i += 1
			except WindowsError:
				pass
		else:
			ins_sharepath = ['/usr/share/tp', 
							 '/usr/share/games/tp', 
							 '/usr/local/share/tp', 
							 '/opt/tp', 
							 os.path.join(version.installpath, 'tp/client/singleplayer'),
				]
		self.build('installed', ins_sharepath)

		# look for inplace single player XML files
		if hasattr(version, 'version_git'):
			inp_sharepath = []
			for repo in [os.path.join('..', r) for r in os.listdir('..')]:
				if os.path.isdir(repo):
					inp_sharepath.append(repo)
			self.build('inplace', inp_sharepath)

	def build(self, stype, sharepath):
		"""\
		Build the local list from single player XML files.

		@param stype: The type of component (installed or inplace).
		@type stype: C{string}
		@param sharepath: A list of paths to search for XML files.
		@type sharepath: C{list} of C{string}
		"""
		for dir in sharepath:
			dir = os.path.abspath(dir)
			if not os.path.isdir(dir):
				continue

			print "Searching in %s..." % dir

			for xmlfile in os.listdir(dir):
				xmlfile = os.path.join(dir, xmlfile)
				if not os.path.isfile(xmlfile):
					continue
				if not xmlfile.endswith('xml'):
					continue

				try:
					xmltree = ET.parse(xmlfile)
				# FIXME: catch actual parsing exceptions (ExpatError?)
				except:
					continue

				# ensure this is a tpconfig document
				if xmltree._root.tag != 'tpconfig':
					continue

				# ensure it is of the type we are looking for
				if (not xmltree._root.attrib.has_key('type') and stype != 'installed') \
				or xmltree._root.attrib.has_key('type') and xmltree._root.attrib['type'] != stype:
					continue

				print "Including %s." % xmlfile
				self.absorb_xml(xmltree, dir)

		# verify existence of command paths referred to in local list
		for t in self.keys():
			for s in self[t].keys():
				exe = os.path.join(self[t][s]['cwd'], self[t][s]['commandstring'].split()[0])
				if not (os.path.exists(exe) or os.path.exists(exe + '.exe')):
					print "Removing %s: command %s not found." % (self[t][s]['longname'], exe)
					del self[t][s]

	def absorb_xml(self, tree, dir, d = None):
		"""\
		Recursively import an XML element tree into the local list. When called
		externally, the tree passed in should be an entire XML file parsed from
		the root, and the d parameter should not be specified. This implementation
		reads documents using the tpconfig DTD, and the classdict classes are the
		major (i.e. with sub-elements) element types specified in that DTD.

		@param tree: The XML element tree to import.
		@type tree: L{ET.ElementTree}
		@param dir: The absolute path of the source XML file.
		@type dir: C{string}
		@param d: A dictionary subclass instance for this type of tree (optional).
		@type d: C{dict}
		"""
		if d is None:
			d = self

		classdict = { 'server' : _Server,
					  'aiclient' : _AIClient,
					  'ruleset' : _Ruleset,
					  'parameter' : _Parameter,
					}

		for k in d.keys():
			if type(d[k]) is dict:
				for s in tree.findall(k):
					sname = s.attrib['name']
					if not d[k].has_key(sname):
						d[k][sname] = classdict[k]()
					self.absorb_xml(s, dir, d[k][sname])
			elif type(d[k]) is list:
				for e in tree.findall(k):
					d[k].append(e.text)
			elif d[k] == '':
				if tree.attrib.has_key(k):
					d[k] = tree.attrib[k]
				elif tree.find(k) is not None:
					d[k] = tree.find(k).text
					if k == 'cwd':
						d[k] = os.path.join(dir, d[k])


class DownloadList(dict):
	"""\
	Builds a list of potentially downloadable servers and AI clients.
	"""

	def __init__(self,
				 urlxml = 'http://thousandparsec.net/tp/downloads.xml',
				 urldlp = 'http://thousandparsec.net/tp/downloads.php'):
		self.urlxml = urlxml
		self.urldlp = urldlp
		self['server'] = {}
		self['ai']  = {}
		self.rulesets = []
		self.got = self.get_list()

	def get_list(self):
		"""\
		Fetch and parse the XML list of available downloads from TP web.

		@return: True if successful, false otherwise.
		@rtype: C{bool}
		"""
		try:
			dlxml = urllib.urlopen(self.urlxml)
			if not dlxml.info()['content-type'] == 'application/xml':
				return False
			xmltree = ET.parse(dlxml)
			for category in xmltree.findall('products/category'):
				cname = category.attrib['name']
				if not cname in self.keys():
					continue

				self[cname] = {}
				for product in category.findall('product'):
					if product.attrib['visible'] == 'no':
						continue

					self[cname][product.attrib['name']] = []
					for rules in product.findall('rules'):
						self[cname][product.attrib['name']].append(rules.text)
						if not rules.text in self.rulesets:
							self.rulesets.append(rules.text)
		# FIXME: Bare excepts are bad.
		except:
			return False
		return True

	def list_servers_with_ruleset(self, rname):
		"""\
		Returns a list of available servers supporting the specified ruleset.

		@param rname: The ruleset name.
		@type rname: C{string}
		@return: A list of servers supporting the ruleset.
		@rtype: C{list} of C{string}
		"""
		servers = []
		for sname in self['server'].keys():
			if rname in self['server'][sname]:
				servers.append(sname)
		return servers

	def list_aiclients_with_ruleset(self, rname):
		"""\
		Returns a list of available AI clients supporting the specified ruleset.

		@param rname: The ruleset name.
		@type rname: C{string}
		@return: A list of AI clients supporting the ruleset.
		@rtype: C{list} of C{string}
		"""
		aiclients = []
		for ainame in self['ai'].keys():
			if rname in self['ai'][ainame]:
				aiclients.append(ainame)
		return aiclients

	def linkurl(self, component = None):
		"""\
		Returns the download page URL, optionally for a specific component type.

		@param component: The component type (optional).
		@type component: C{string}
		@return: A download URL (for component type or all).
		@rtype: C{string}
		"""
		if component in self.keys():
			return self.urldlp + '?category=' + component
		else:
			return self.urldlp

class InitError(Exception):
	"""\
	Generic initialization error, thrown to allow cleanup in certain situations.
	"""
	pass

class SinglePlayerGame:
	"""\
	The single-player game manager. This is the object which should be
	instantiated externally to create a single player game.
	"""

	def __init__(self):
		# build local list
		self.locallist = LocalList()

		# initialize internals
		self.active = False
		self.sname = ''
		self.rname = ''
		self.sparams = {}
		self.rparams = {}
		self.opponents = []

	def __del__(self):
		if self.active:
			self.stop()

	@property
	def servers(self):
		"""\
		Returns a list of available servers.

		@return: A list of servers.
		@rtype: C{list} of C{string}
		"""
		return self.locallist['server'].keys()
	
	@property
	def aiclients(self):
		"""\
		Returns a list of available AI clients.

		@return: A list of AI clients.
		@rtype: C{list} of C{string}
		"""
		return self.locallist['aiclient'].keys()

	@property
	def rulesets(self):
		"""\
		Returns a list of available rulesets from all servers.

		@return: A list of rulesets.
		@rtype: C{list} of C{string}
		"""
		rulesets = []
		for sname in self.locallist['server'].keys():
			for rname in self.locallist['server'][sname]['ruleset'].keys():
				if rname not in rulesets:
					rulesets.append(rname)
		rulesets.sort()
		return rulesets

	def server_info(self, sname = None):
		"""\
		Returns information about a server.

		@param sname: Server name (optional).
		@type sname: C{string}
		@return Information about current or specified server.
		@rtype: C{dict}
		"""
		if sname is None:
			sname = self.sname
		try:
			return self.locallist['server'][sname]
		except KeyError, e:
			return None

	def aiclient_info(self, ainame = None):
		"""\
		Returns information about an AI client.

		@param ainame: AI client name.
		@type ainame: C{string}
		@return: Information about specified AI client.
		@rtype: C{dict}
		"""
		try:
			return self.locallist['aiclient'][ainame]
		except KeyError, e:
			return None

	def ruleset_info(self, rname = None):
		"""\
		Returns information about a ruleset.

		@param rname Ruleset name (optional).
		@return Information about current or specified ruleset.
		"""
		if rname is None:
			rname = self.rname
		if self.sname:
			sname = self.sname
		else:
			for sname in self.locallist['server'].keys():
				if self.locallist['server'][sname]['ruleset'].has_key(rname):
					break
		try:
			return self.locallist['server'][sname]['ruleset'][rname]
		except KeyError, e:
			return None

	def list_servers_with_ruleset(self, rname = None):
		"""\
		Returns a list of servers supporting the current or specified ruleset.

		@param rname: Ruleset name (optional).
		@type rname: C{string}
		@return: A list of servers.
		@rtype: C{list} of C{string}
		"""
		if rname is None:
			rname = self.rname
		servers = []
		for sname in self.locallist['server'].keys():
			if self.locallist['server'][sname]['ruleset'].has_key(rname):
				servers.append(sname)
		servers.sort()
		return servers

	def list_aiclients_with_ruleset(self, rname = None):
		"""\
		Returns a list of AI clients supporting the current or specified ruleset.

		@param rname: Ruleset name (optional).
		@type rname: C{string}
		@return: A list of AI clients.
		@rtype: C{list} of C{string}
		"""
		if rname is None:
			rname = self.rname
		aiclients = []
		for ainame in self.locallist['aiclient'].keys():
			if rname in self.locallist['aiclient'][ainame]['rules']:
				aiclients.append(ainame)
		aiclients.sort()
		return aiclients

	def list_sparams(self, sname = None):
		"""\
		Returns the parameter list for the current or specified server.

		@param sname: Server name (optional).
		@type sname: C{string}
		@return: The server parameter list.
		@rtype: C{dict}
		"""
		if sname is None:
			sname = self.sname
		return self.locallist['server'][sname]['parameter']

	def list_aiparams(self, ainame):
		"""\
		Returns the parameter list for the specified AI client.

		@param ainame: AI client name.
		@type ainame: C{string}
		@return: The AI client parameter list.
		@rtype: C{dict}
		"""
		return self.locallist['aiclient'][ainame]['parameter']

	def list_rparams(self, sname = None, rname = None):
		"""\
		Returns the parameter list for the current or specified ruleset.

		@param rname: Ruleset name (optional).
		@type rname: C{string}
		@return: The ruleset parameter list.
		@rtype: C{dict}
		"""
		if sname is None:
			sname = self.sname
		if rname is None:
			rname = self.rname
		return self.locallist['server'][sname]['ruleset'][rname]['parameter']
	
	def add_opponent(self, name, user, parameters):
		"""\
		Adds an AI client opponent to the game (before starting).

		@param name: The name of the AI client.
		@type name: C{string}
		@param user: The desired username of the opponent.
		@type user: C{string}
		@param parameters: A dictionary of parameters in the form {'name', 'value'}.
		@type parameters: C{dict}
		@return: True if successful, false otherwise.
		@rtype: C{bool}
		"""
		for aiclient in self.opponents:
			if aiclient['user'] is user:
				return False

		aiclient = {
				'name' : name,
				'user' : user,
				'parameters' : parameters,
			}
		self.opponents.append(aiclient)

		return True

	def start(self):
		"""\
		Starts the server and AI clients.

		@return: Port number (OK to connect) or False.
		@rtype: C{int}
		"""
		import atexit
		atexit.register(self.stop)

		if self.active:
			return

		# find a free port
		s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		s.bind(('localhost',0))
		port = s.getsockname()[1]
		s.close()

		try:
			# start server
			server = self.locallist['server'][self.sname]
			ruleset = server['ruleset'][self.rname]

			# start server - create server command line
			servercmd = server['commandstring']
			
			# start server - set working directory
			servercwd = os.path.normpath(server['cwd'])
			if servercwd == '':
				servercwd = None
			else:
				servercmd = os.path.join(servercwd, servercmd)

			# start server - add forced parameters to command line
			for forced in server['forced']:
				servercmd += ' ' + forced

			# start server - set ruleset and port
			servercmd = servercmd % {
						'rname' : self.rname,
						'port' : port,
					}

			# start server - add regular parameters to command line
			for pname in server['parameter'].keys():
				value = server['parameter'][pname]['default']
				if self.sparams.has_key(pname):
					value = self.sparams[pname]
				value = self._format_value(value, server['parameter'][pname]['type'])
				if value is None:
					continue
				servercmd += ' ' + server['parameter'][pname]['commandstring'] % value

			# start server - add forced ruleset parameters to command line
			for forced in ruleset['forced']:
				servercmd += ' ' + forced
			
			# start server - add regular ruleset parameters to command line
			for pname in ruleset['parameter'].keys():
				value = ruleset['parameter'][pname]['default']
				if self.rparams.has_key(pname):
					value = self.rparams[pname]
				value = self._format_value(value, ruleset['parameter'][pname]['type'])
				if value is None:
					continue
				elif value == '':
					servercmd += ' ' + ruleset['parameter'][pname]['commandstring']
				else:
					servercmd += ' ' + ruleset['parameter'][pname]['commandstring'] % value

			# start server - call the control script
			# TODO: allow redirection of stdout and stderr
			try:
				self.sproc = Popen(servercmd.split(), cwd = servercwd)
			except OSError, e:
				raise InitError(e)

			print "Running server with cmd:", servercmd

			# wait for the server to initialize
			# FIXME: use admin protocol if available to check this (loop)
			time.sleep(5)
	
			# start AI clients
			for aiclient in self.opponents:
				# create ai client command line
				aicmd = self.locallist['aiclient'][aiclient['name']]['commandstring']

				# set working directory
				aicwd = os.path.normpath(self.locallist['aiclient'][aiclient['name']]['cwd'])
				if aicwd == '':
					aicwd = None
                                else:
                                        aicmd = os.path.join(aicwd, aicmd)

				# add forced parameters to command line
				for forced in self.locallist['aiclient'][aiclient['name']]['forced']:
					aicmd += ' ' + forced

				# set port, ruleset and username
				aicmd = aicmd % {
						'port' : port,
						'rname' : self.rname,
						'user' : aiclient['user'],
					}

				# add regular parameters to command line
				for pname in self.locallist['aiclient'][aiclient['name']]['parameter'].keys():
					value = self.locallist['aiclient'][aiclient['name']]['parameter'][pname]['default']
					if aiclient['parameter'].has_key(pname):
						value = aiclient['parameter'][pname]
					value = self._format_value(value, self.locallist['aiclient'][aiclient['name']]['parameter'][pname]['type'])
					if value is None:
						continue
					aicmd += ' ' + self.locallist['aiclient'][aiclient['name']]['parameter'][pname]['commandstring'] % value

				print "Running AI (%s) with cmd: %s" % (aiclient['name'], aicmd)

				# call the control script
				# TODO: allow redirection stdout and stderr
				try:
					aiclient['proc'] = Popen(aicmd.split(), cwd = aicwd)
				except OSError, e:
					raise InitError(e)

			# set active flag
			self.active = True

			# ensure that processes stay alive
			time.sleep(2)
			if not self.sproc.poll() is None:
				raise InitError
			for aiclient in self.opponents:
				if not aiclient['proc'].poll() is None:
					raise InitError

		except InitError, e:
			print e
			self.stop()
			return False

		return port

	def stop(self):
		"""\
		Stops the server and AI clients.
		Should be called by the client when disconnecting/closing.
		"""
		if not self.active:
			return

		# stop server
		if self.sname != '':
			try:
				self.sproc.kill()
			except OSError, e:
				pass
			self.sname = ''
			self.rname = ''

		# stop AI clients
		for aiclient in self.opponents:
			try:
				aiclient['proc'].kill()
			except OSError, e:
				pass
		self.opponents = []

		# reset active flag
		self.active = False

	def _format_value(self, value, ptype):
		"""\
		Internal: formats a parameter value based on type.

		@oaram value: The value to format.
		@type value: C{string}
		@param ptype: The target value type (I, S, F, B).
		@type ptype: C{string}
		@return: The formatted value or None.
		"""
		if value is None or str(value) == '':
			return None
		elif ptype == 'I':
			return int(value)
		elif ptype == 'S' or type == 'F':
			return str(value)
		elif ptype == 'B' and str(value) == 'True':
			return ''
		else:
			return None



#!/usr/bin/env python

import sys

try:
	del sys.argv[sys.argv.index('--ignore-deps')]
	print "Ignoring dependencies..."
except ValueError:
	import pkg_resources
	pkg_resources.require('libtpproto-py')

from tp.client import __version__
version = "%s.%s.%s" % __version__[0:3]

from setuptools import setup

setup(
	name		="libtpclient-py",
	version		=version,
	license		="GPL",
	description	="Client support library for Thousand Parsec",
	long_description="""\
A library of code to support quick development of Clients for Thousand Parsec.

Includes support for:
	* Classes of keeping a download cache of the universe (including automatic
 update)
	* Classes for parsing and calculating tpcl 
	* Threading support
""",
	author		="Tim Ansell",
	author_email="tim@thousandparsec.net",
	url			="http://www.thousandparsec.net",
	keywords	="thousand parsec space client support empire building strategy game tpcl scheme",

	namespace_packages = ['tp'],
	packages=[ \
		'tp',
		'tp.client',
		'tp.client.pyscheme',
	],
	zip_safe=False,

	include_patckage_data = True,
	package_data = { 'tp.client': ['singleplayer/*.dtd', 'singleplayer/aiclients/*', 'singleplayer/servers/*'] }
)




from version import version as vi
from version import installpath as ip
__version__     = vi
__installpath__ = ip

import __builtin__
try:
	__builtin__._
except AttributeError:
	def _(s):
		return s
	__builtin__._ = _




"""parser.py --- A small scheme parser.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

This parser might be useful if we wanted to grab information from the
user using Schemeish lists.  It might also be nice if one is planning
to write a Scheme interpreter in Python.  *cough*


The main functions to use in this module are

    tokenize(): makes a list of tokens out of a string

    parse(s): parse string s, and return a either an atomic value
        (symbol or number) or a pair.

If anything bad happens during the parsing, we'll raise a ParserError.


Some special notes: this module should behave well even under unusual
input: there should be no possibility of hitting the recursion limit,
since we are using trampolined style.
"""

__license__ = "MIT License"

import re
from symbol import Symbol
import pair
import pogo


"""The End Of File token is a sentinal that terminates a list of tokens."""
EOF_TOKEN = (None, None)


"""Here are the patterns we pay attention when breaking down a string
into tokens."""
PATTERNS = [ ('whitespace', re.compile(r'(\s+)')),
             ('comment', re.compile(r'(;[^\n]*)')),
             ('(', re.compile(r'(\()')),
             (')', re.compile(r'(\))')),
             ('number', re.compile(r'''( [+\-]?    ## optional sign,
                                         (?:       ## followed by some
                                                   ## decimals
                                              \d+\.\d+
                                            | \d+\.
                                            | \.\d+
                                            | \d+
                                         )
                                       )
                                         ''',
                                   re.VERBOSE)),
             ('symbol',
              re.compile(r'''([a-zA-Z\+\=\?\!\@\#\$\%\^\&\*\-\/\.\>\<]
                              [\w\+\=\?\!\@\#\$\%\^\&\*\-\/\.\>\<]*)''',
                                   re.VERBOSE)),
             ('string', re.compile(r'''
                                      "
                                      (([^\"] | \\")*)
                                      "
                                      ''',
                                   re.VERBOSE)),
             ('\'', re.compile(r'(\')')),
             ('`', re.compile(r'(`)')),
             ## fixme: add UNQUOTE-SPLICING form as well
             (',', re.compile(r'(,)')),
             ]


def tokenize(s):
    """Given a string 's', return a list of its tokens.  A token can
    be one of the following types listed in the PATTERNS above, and
    each token is a 2-tuple: (type, content)."""
    tokens = []
    while 1:
        should_continue = 0
        for tokenType, regex in PATTERNS:
            match_obj = regex.match(s)
            if match_obj:
                should_continue = 1
                tokens.append( (tokenType, match_obj.group(1)) )
                s = s[match_obj.span()[1] :]
        if should_continue == 0:
            break
    tokens.append(EOF_TOKEN)
    return filter(lambda x: x[0] not in ('whitespace', 'comment'),
                  tokens)


"""With the lexer done, now let's direct our attention to the parser."""
######################################################################

class ParserError(Exception):
    """Our personalized exception class."""
    pass


def peek(tokens):
    """Take a quick glance at the first token in our tokens list."""
    if len(tokens) == 0:
        raise ParserError, "While peeking: ran out of tokens."
    return tokens[0]


def eat(tokens, tokenType):
    """Digest the first token in our tokens list, making sure that we're
    biting on the right tokenType of thing."""
    if len(tokens) == 0:
        raise ParserError, "While trying to eat %s: ran out of tokens." % \
              (repr(tokenType),)
    if tokens[0][0] != tokenType:
        raise ParserError, "While trying to eat %s: got token %s instead." % \
                            (repr(tokenType), repr(tokens[0]))
    return tokens.pop(0)

######################################################################




def identity_cont(val):
    return pogo.land(val)


"""Below is our parser for symbol-lists.  We'll use a recursive
descent parsing technique, since Scheme's syntax is so well suited for
it.  Actually, it's a bit more complicated than this.

There is one major problem with a naive rec-descent approach: with
unusual input, we can run into Python's recursion stack limit.  To get
around this, we first wrote the parser normally, and then attacked the
recursion using CPS and trampolining style.  The result is that the
parser works, even with evil input, but it's a bit harder to read.
"""


def parseSingleExpression(tokens, cont=identity_cont):
    """Returns a single Expression, given a sequence of tokens.
    Raises a ParserException if our tokens haven't been exhausted."""
    def c(expression):
        eat(tokens, None)
        return pogo.bounce(cont, expression)

    return parseExpression(tokens, c)


def parseExpression(tokens, cont=identity_cont):
    """Returns an Expression, given a sequence of tokens.
    An expression is made up of one of the following things:
        o  A quoted expression
        o  An atom (like a number or symbol or string)
        o  A list.
    This procedure tries to take care of all these potentials."""
    look_ahead_type = peek(tokens)[0]

    def make_c(quoteType):
        def c(expr):
            return pogo.bounce(cont, pair.list(Symbol(quoteType), expr))
        return c
    if look_ahead_type == '\'':
        eat(tokens, '\'')
        return parseExpression(tokens, make_c('quote'))
    if look_ahead_type == '`':
        eat(tokens, '`')
        return parseExpression(tokens, make_c('quasiquote'))
    elif look_ahead_type == ',':
        eat(tokens, ',')
        return parseExpression(tokens, make_c('unquote'))
    elif look_ahead_type == '(':
        return parseList(tokens, cont)
    elif look_ahead_type in ('number', 'symbol', 'string'):
        return parseAtom(tokens, cont)
    else:
        raise ParserError, "While parsing Expression: no alternatives."


def parseAtom(tokens, cont):
    """Returns an Atom, given a sequence of tokens.
    An atom is either a number, a symbol, or a string."""
    if peek(tokens)[0] == 'number':
        return pogo.bounce(cont, toNumber(eat(tokens, 'number')[1]))
    elif peek(tokens)[0] == 'symbol':
        return pogo.bounce(cont, Symbol(eat(tokens, 'symbol')[1]))
    elif peek(tokens)[0] == 'string':
        return pogo.bounce(cont, eat(tokens, 'string')[1])
    else:
        raise ParserError, "While parsing Atom: no alternatives."


def toNumber(s):
    """Tries to convert string 's' into a number."""
    try:
        return int(s)
    except ValueError:
        return float(s)





def parseList(tokens, cont):
    """Parses a parenthesized list expression."""
    eat(tokens, "(")
    def c_expressionsEaten(val):
        eat(tokens, ")")
        return pogo.bounce(cont, val)
    return pogo.bounce(parseExpressionStar, tokens, c_expressionsEaten)


def parseExpressionStar(tokens, cont):
    """Tries to eat as many expressions as it can see."""
    START_EXPR_SET = ('\'', '`', ',', '(', 'number', 'symbol', 'string')
    if peek(tokens) == ('symbol', '.'):
        ## Handle dotted pairs
        eat(tokens, "symbol")
        return pogo.bounce(parseExpression, tokens, cont)
    elif peek(tokens)[0] in START_EXPR_SET:
        def c_first_eaten(firstVal):
            def c_rest_eaten(restVal):
                return pogo.bounce(cont, pair.cons(firstVal, restVal))
            return pogo.bounce(parseExpressionStar, tokens, c_rest_eaten)
        return pogo.bounce(parseExpression, tokens, c_first_eaten)
    else:
        return pogo.bounce(cont, pair.NIL)


def parse(s):
    """Parse a single string.  This is just a convenience function."""
    return pogo.pogo(parseSingleExpression(tokenize(s),
                                           identity_cont))



######################################################################

try:
    import unittest
    class ParserTests(unittest.TestCase):
        def testAtomParsing(self):
            self.assertEquals(42, parse("42"))
            self.assertEquals("particle-man", parse('"particle-man"'))


        def testBrokenNegativeExampleFromBaypiggies(self):
            ## embarassing case that didn't work during the Baypiggies meeting
            ## on September 9, 2004.  Doh!
            self.assertEquals(-42, parse("-42")) 
            

        def testLists(self):
            self.assertEquals(pair.list(1, 2), parse("(1 2)"))
            self.assertEquals(pair.list(1, 2, pair.list(3), 4),
                              parse("(1 2 (3) 4)"))
            self.assertEquals(pair.list(pair.list(pair.list())),
                                              parse("((()))"))

        def testQuotation(self):
            self.assertEquals(pair.list(Symbol("quote"),
                                              Symbol("atom-man")),
                              parse("'atom-man"))
            self.assertEquals(pair.list(Symbol("quasiquote"),
                                              Symbol("istanbul")),
                              parse("`istanbul"))

            self.assertEquals(pair.list(Symbol("unquote"),
                                              Symbol("constantinople")),
                              parse(",constantinople"))

        def testDottedPair(self):
            cons = pair.cons ## shortcut
            self.assertEquals(cons(Symbol("alpha"), Symbol("beta")),
                              parse("(alpha . beta)"))
            self.assertEquals(cons(Symbol("a"),
                                   cons(Symbol("b"),
                                        cons(cons(Symbol("c"), Symbol("d")),
                                             pair.NIL))),
                              parse("(a b (c . d))"))


        def testQuotedList(self):
            self.assertEquals(pair.list(Symbol("quote"),
                                              pair.list(Symbol("foo"),
                                                              Symbol("bar"))),
                              parse("'(foo bar)"))


        def testEmptyList(self):
            self.assertEquals(pair.NIL, parse("()"))


        def testStressWithSuperNesting(self):
            ## An evil test to see if this raises bad errors.
            N = 1000
            bigNestedList = pair.list()
            for ignored in xrange(N-1):
                bigNestedList = pair.list(bigNestedList)
            try:
                self.assertEquals(bigNestedList,
                                  parse( "(" * N + ")" * N))
            except RuntimeError, e:
                self.fail(e)


    if __name__ == '__main__':
        unittest.main()
except ImportError:
    pass



"""Defines a base built-in environment, with all the primitives."""

__license__ = "MIT License"

import types
import math
import sys

import pogo
from symbol import Symbol, true, false
import expressions
import environment
import pair
from parser import parse
import evaluator
from error import SchemeError
import operator




def schemeParse(string):
    """Given a string, returns the pair list parsing of that string."""
    return parse(string)


def allNumbers(numbers):
    for n in numbers:
        if type(n) not in (types.IntType, types.LongType, types.FloatType):
            return 0
    return 1


def schemeAdd(*numbers):
    if len(numbers) == 0: return 0
    if not allNumbers(numbers):
        raise SchemeError, "primitive '+' --- arguments must be numbers"
    return reduce(operator.add, numbers)

def schemeSubtract(*numbers):
    if len(numbers) == 0:
        raise SchemeError, "primitive '-' --- expects 1 or more arguments"
    if not allNumbers(numbers):
        raise SchemeError, "primitive '-' --- arguments must be numbers"
    if len(numbers) == 1:
        return -numbers[0]
    return reduce(operator.sub, numbers)

def schemeMultiply(*numbers):
    if len(numbers) == 0: return 1
    if not allNumbers(numbers):
        raise SchemeError, "primitive '*' --- arguments must be numbers"
    return reduce(operator.mul, numbers)

def schemeDivide(*numbers):
    if len(numbers) == 0:
        raise SchemeError, "primitive '/' --- expects 1 or more arguments"
    if not allNumbers(numbers):
        raise SchemeError, "primitive '/' --- arguments must be numbers"
    return reduce(operator.div, numbers)

def schemeRemainder(a, b):
    if not allNumbers([a, b]):
        raise SchemeError, "primitive 'remainder' --- arguments must be numbers"
    return a % b

def schemePower(a, b):
    if not allNumbers([a, b]):
        raise SchemeError, "primitive 'power' --- arguments must be numbers"
    return a ** b

def schemeQuotient(a, b):
    if not allNumbers([a, b]):
        raise SchemeError, "primitive 'quotient' --- arguments must be numbers"
    return int(a / b)

def schemeSqrt(a):
    if not allNumbers([a]):
        raise SchemeError, "primitive 'sqrt' --- argument must be a number"
    return math.sqrt(a)

def schemeFloor(a):
    if not allNumbers([a]):
        raise SchemeError, "primitive 'floor' --- argument must be a number"
    return int(math.floor(a))

def schemeCeiling(a):
    if not allNumbers([a]):
        raise SchemeError, "primitive 'ceiling' --- argument must be a number"
    return int(math.ceil(a))

def schemeNumericalEq(a, b):
    """Here, we need to make sure that we're returning the metacircular
    values of true or false."""
    return schemeBooleanize(a == b)

def schemeLessThan(a, b): return schemeBooleanize(a < b)
def schemeLessThanOrEquals(a, b): return schemeBooleanize(a <= b)
def schemeGreaterThan(a, b): return schemeBooleanize(a > b)
def schemeGreaterThanOrEquals(a, b): return schemeBooleanize(a >= b)

def schemeEqQuestion(a, b):
    if type(a) in (types.IntType, types.FloatType, types.LongType):
        return schemeBooleanize(a == b)
    return schemeBooleanize(a is b)

def schemeEqualQuestion(a, b): return schemeBooleanize(a == b)
def schemeListQuestion(a): return schemeBooleanize(pair.isList(a))
def schemeNullQuestion(a): return schemeBooleanize(pair.isNull(a))


def schemeDisplay(thing):
    sys.stdout.write(expressions.toString(thing,
                                          quoteStrings=0))
    return Symbol("ok")


def schemeWrite(thing):
    sys.stdout.write(expressions.toString(thing,
                                          quoteStrings=1))
    return Symbol("ok")


def schemeNewline():
    sys.stdout.write("\n")
    return Symbol("ok")

def schemeStringToSymbol(mystr): return Symbol(mystr)
def schemeSymbolToString(mysymbol): return str(mysymbol)
def schemeNumberToString(mynumber): return str(mynumber)

def schemeStringAppend(*strings):
    if len(strings) == 0: return ""
    return reduce(operator.add, strings)


def schemeNot(x):
    if x == false: return true
    return false

def schemeQuit():
    sys.stdout.write("bye\n")
    sys.exit(0)

def schemeBooleanize(x):
    """Translates what Python considers true and false to the Scheme
    symbols "true" or "false"."""
    if x: return true
    return false


def schemePairQuestion(L):
    return schemeBooleanize(pair.isPair(L))


def schemeError(*args):
    raise SchemeError, ' '.join(map(expressions.toString, args))


def wrapPrimitiveWithDefaults(f):
    """Normal primitives don't know what to do with their continuation
    or environment arguments, so this wrapper provides a nice
    default."""
    def newPrimitive(cont, env, args):
        return pogo.bounce(cont, (f(*args)))
    return newPrimitive


######################################################################


def installPythonFunction(name, function, env, wrapDefaults=True):
    """Installs a new Python function as a primitive into the given
    environment 'env'"""
    if wrapDefaults:
        wrappedProcedure = expressions.makePrimitiveProcedure(
            wrapPrimitiveWithDefaults(function))
    else:
        wrappedProcedure = expressions.makePrimitiveProcedure(function)
    environment.defineVariable(Symbol(name), wrappedProcedure,
                               env)


def setupEnvironment():
    """Sets up a new environment with a bunch of fairly standard
    Scheme built-in primitives."""
    PRIMITIVE_PROCEDURES = [
        ["car", pair.car],
        ["cdr", pair.cdr],
        ["cons", pair.cons],
        ["append", pair.append],
        ["list", pair.list],
        ["set-car!", pair.setCarBang],
        ["set-cdr!", pair.setCdrBang],
        ["+", schemeAdd],
        ["-", schemeSubtract],
        ["*", schemeMultiply],
        ["/", schemeDivide],
        ["remainder", schemeRemainder],
        ["quotient", schemeQuotient],
        ["sqrt", schemeSqrt],
        ["floor", schemeFloor],
        ["ceiling", schemeCeiling],
        ["=", schemeNumericalEq],
        ["<", schemeLessThan],
        ["<=", schemeLessThanOrEquals],
        [">", schemeGreaterThan],
        [">=", schemeGreaterThanOrEquals],
        ["eq?", schemeEqQuestion],
        ["equal?", schemeEqualQuestion],
        ["list?", schemeListQuestion],
        ["pair?", schemePairQuestion],
        ["null?", schemeNullQuestion],
        ["display", schemeDisplay],
        ["write", schemeWrite],
        ["newline", schemeNewline],
        ["not", schemeNot],
        ["string->symbol", schemeStringToSymbol],
        ["symbol->string", schemeSymbolToString],
        ["number->string", schemeNumberToString],
        ["string-append", schemeStringAppend],
        ["quit", schemeQuit],
        ["exit", schemeQuit],
        ["error", schemeError],
        ["parse", schemeParse],
        ["pow", schemePower],
        ]

    initial_environment = environment.extendEnvironment(
        pair.NIL, pair.NIL, environment.THE_EMPTY_ENVIRONMENT)

    for name, proc in PRIMITIVE_PROCEDURES:
        installPythonFunction(name, proc, initial_environment)


    ## Finally, put true and false in there.
    environment.defineVariable(Symbol("#t"), Symbol("#t"), initial_environment)
    environment.defineVariable(Symbol("#f"), Symbol("#f"), initial_environment)
    return initial_environment



"""A small prompting class.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

This hasn't been too documented yet.  There's some sample uses on
the bottom with the test() function below.

Slightly modified to work nicely with parentheses.
"""

__license__ = "MIT License"

## First, let's see if we can load up readline and make raw_input()
## nice to work with.
try:
    import readline
except ImportError: pass

import sys
import pprint

class Prompt:
    def __init__(self, name, quit_str = 'quit', quit_cmd=sys.exit,
                 callback=None):
        self.name = name
        self.quit_str = quit_str
        self.quit_cmd = quit_cmd
        self.callback = callback
        self.s = ''

    def makePrompt(self):
        if self.countOpenParens():
            return '[%s%d)] >>> ' % \
                   ('.' * (len(self.name)
                           - len(str(self.countOpenParens()))
                           - 1),
                    self.countOpenParens())
        else:
            return '[%s] >>> ' % self.name

    def countOpenParens(self):
        """Returns the number of open parens in self.s"""
        return self.s.count('(') - self.s.count(')')
        
    def promptLoop(self):
        while 1:
            try:
                self.s = self.s + '\n' + raw_input(self.makePrompt())
                if self.s == self.quit_str:
                    raise EOFError
                elif self.countOpenParens() > 0:
                    pass
                elif self.s.strip() and self.callback:
                    result = self.callback(self.s)
                    if result != None:
                        pprint.pprint(result)
                    self.s = ''
            except EOFError:
                print
                return self.quit_cmd()
            except KeyboardInterrupt:
                self.s = ''
                print



######################################################################

## Note: I should probably use the unit testing modules.
def test():
    def echoCallback(s):
        print s
    p = Prompt('test', callback=echoCallback)
    p.promptLoop()


if __name__ == '__main__':
    test()




version = (0, 3, 2)

# Add the git version if in a git tree...
import os, os.path
__path__ = os.path.realpath(os.path.dirname(__file__))
installpath = os.path.split(os.path.split(__path__)[0])[0]

# Get the git version this tree is based on
if os.path.exists(os.path.join(installpath, '.git')):
	# Read in git's 'HEAD' file which points to the correct reff to look at
	h = open(os.path.join(installpath, '.git', 'HEAD')).readline().strip()
	try:
		# Read in the ref
		ref = h.split(': ', 1)[1]
		# This file has the SHA1
		p = open(os.path.join(installpath, '.git', ref))
		del ref

		version_git = p.read().strip()
	except IndexError:
		version_git = h

	# What version are we trying to get too?
	import time
	if version[2] >= 99:
		version_target = (version[0], version[1]+1, 0)
	else:
		version_target = (version[0], version[1], version[2]+1)

	version_target_str = "%i.%i.%i" % version_target

version_str = "%i.%i.%i" % version[:3]

if __name__ == "__main__":
	import sys
	if len(sys.argv) > 1 and sys.argv[1] == '--fix':
		print """
import os, os.path
__path__ = os.path.realpath(os.path.dirname(__file__))
installpath = os.path.split(os.path.split(__path__)[0])[0]
"""
		for value in dir():
			if value.startswith('__') or value in ('installpath',):
				continue
			if isinstance(eval(value), (tuple, str, unicode)):
				exec("print '%s =', repr(%s)" % (value, value))

		print """
try:
	print version_str+'+'+version_target_str, "(git %s)" % version_git, "(installed at %s)" % installpath
except ValueError:
	print version_str, "(installed at %s)" % installpath
"""
		sys.exit(0)

	if os.path.exists(os.path.join(installpath, '.git')):
		print version_str+'+'+version_target_str, "(git %s)" % version_git, "(installed at %s)" % installpath
	else:
		print version_str, "(installed at %s)" % installpath



"""Implementation of pairs for Python.

Pairs are like linked lists in Python.  For historical reasons, we use
the following funny names to construct and destructure lists:

    cons(head, tail) --- constructs a ConsPair that consists of a
    head and a tail.

    car(pair) --- returns the head of a ConsPair.

    cdr(pair) --- returns the tail of a ConsPair.

Scheme's lists are built up of pair chains terminated by NIL.

The reason we use a separate factory function --- cons() --- instead
of directly using the ConsPair is because we may want the freedom to
change implementation later on.  In SICP, in fact, there's a totally
screwy implementation that uses lambdas entirely, with no real data
structure.

There are a few more convenience functions here to rapidly make these
pair structures.  For example, there's a list() function here that,
given a set of arguments, produces a pair chain.
"""


__license__ = "MIT License"


"""This module is not 'from pair import *' safe!  Particularly because
we define a few functions here that have names that conflict with
builtins.  Let's enforce this restriction."""
__all__ = []


from sets import Set

from symbol import Symbol
from error import SchemeError
import pogo


"""The NIL atom.  There should be only one!"""
NIL = []


class ConsPair(list): pass


def isNull(x):
    return x is NIL


def isPair(p):
    return type(p) is ConsPair


def cons(head, rest):
    """Returns the concatentation of head with rest."""
    return ConsPair([head, rest])



def car(p):
    """Returns the head of the pair."""
    if not isPair(p):
        raise SchemeError, "CAR --- argument is not a pair."
    if p is NIL:
        raise SchemeError, "CAR --- cannot take CAR of empty list."
    else:
        return p[0]


def cdr(p):
    """Returns the tail of the pair."""
    if not isPair(p):
        raise SchemeError, "CDR --- argument is not a pair."
    if p is NIL:
        raise SchemeError, "CDR --- cannot take CDR of empty list."
    else:
        return p[1]


def cddr(p):
    return cdr(cdr(p))

def cdddr(p):
    return cdr(cdr(cdr(p)))


def cadr(p):
    return car(cdr(p))

def caddr(p):
    return car(cdr(cdr(p)))

def cadddr(p):
    return car(cdr(cdr(cdr(p))))



def setCarBang(pair, element):
    """Sets the head of the pair to the element."""
    if not isPair(pair):
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[0] = element
    return Symbol("ok")


def setCdrBang(pair, element):
    """Sets the tail of the pair to the element."""
    if not isPair(pair):
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[1] = element
    return Symbol("ok")



## Hmm... some of these functions really belong in builtins.py.

def isList(p):
    """Returns True if p refers to a list-like structure.
    Note: loopy structures don't qualify as lists.
    """
    seenPairIds = Set()
    while True:
        if id(p) in seenPairIds:
            return 0
        seenPairIds.add(id(p))
        if isNull(p): return 1
        if not isPair(p): return 0
        p = cdr(p)


def isDottedPair(p):
    """Returns True if p refers to an improper list, where the cdr is
    not a pair."""
    if not isPair(p): return 0
    elif isPair(cdr(p)) or isNull(cdr(p)): return 0
    else: return 1


def length(p):
    """Returns the length of p.  Assumes that p is a list."""
    if not isList(p): raise SchemeError, "LENGTH --- not a list"
    length = 0
    while True:
        if isNull(p): return length
        length += 1
        p = cdr(p)


def reverse(p):
    """Reverses a list."""
    if not isList(p): raise SchemeError, "REVERSE --- not a list"
    result = NIL
    while not isNull(p):
        result = cons(car(p), result)
        p = cdr(p)
    return result



def listMap(f, p):
    """Maps a function f across p."""
    if not isList(p): raise SchemeError, "MAP --- not a list"
    resultsRev = NIL
    while not isNull(p):
        resultsRev = cons(f(car(p)), resultsRev)
        p = cdr(p)
    return reverse(resultsRev)



def c_listMap(c_f, p, cont, allow_improper_lists=False):
    """Maps a function f across p, but in a continuation-passed style.

    'c_f' is a function that takes a 2-tuple (element, cont),
    the element and the continuation.

    'cont' is the continutation we apply on the mapped list.

    If the optional keyword parameter 'allow_improper' is set to True,
    then we'll also allow mapping across improper lists.
    """
    if isNull(p):
        return pogo.bounce(cont, NIL)
    if not isPair(p):
        if allow_improper_lists:
            return pogo.bounce(c_f, p, cont)
        else:
            raise SchemeError, "CMAP --- not a list"
    def c_head(head_val):
        def c_tail(tail_val):
            return pogo.bounce(cont, cons(head_val, tail_val))
        return pogo.bounce(c_listMap, c_f, cdr(p), c_tail,
                           allow_improper_lists)
    return pogo.bounce(c_f, car(p), c_head)


    
def append(*lists):
    """Appends all lists together."""
    appended_list = lists[-1]
    for i in xrange(len(lists)-2, -1, -1):
        next_list = lists[i]
        appended_list = pogo.pogo(
            c_appendTwo(next_list, appended_list, pogo.land))
    return appended_list


def c_appendTwo(front, back, cont):
    """Appends two lists together.  Written in continuation passing style."""
    if not isList(front): raise SchemeError, "MAP --- not a list"
    if not isList(back): raise SchemeError, "MAP --- not a list"
    if isNull(front):
        return pogo.bounce(cont, back)
    def c(appendedRest):
        return pogo.bounce(cont, cons(car(front),
                                                appendedRest))
    return pogo.bounce(c_appendTwo, cdr(front), back, c)



def toPythonList(pair):
    """Does a shallow conversion of a pair list chain to a Python list."""
    if not isList(pair): raise SchemeError, "not a list"
    elements = []
    while not isNull(pair):
        elements.append(car(pair))
        pair = cdr(pair)
    return elements


"""Let's save the old version of Python's list function."""
list_in_underlying_python = list 


def list(*elements):
    """Does a shallow conversion of a Python list to a pair chain.

Warning: this does have the same name as the builtin list() function
in Python.
    """
    result = list_in_underlying_python(elements)
    result.reverse()
    return reduce(lambda x, y: cons(y, x), [NIL] + result)


######################################################################

try:
    import unittest
    class PairTests(unittest.TestCase):
        def testNull(self):
            self.assert_(isList(NIL))

        def testReversal(self):
            self.assertEquals(list(1, 2, 3, 4, 5),
                              reverse(list(5, 4, 3, 2, 1)))

        def testMapping(self):
            self.assertEquals(list(2, 4, 6, 8),
                              listMap(lambda x: x*2, list(1, 2, 3, 4)))

        def testAppendTwo(self):
            self.assertEquals(list(1, 2, 3, 4),
                              pogo.pogo(c_appendTwo(list(1, 2),
                                                    list(3, 4),
                                                    pogo.land)))


        def testContinuationMapping(self):
            def c_square(x, cont):
                return pogo.bounce(cont, x**2)

            self.assertEquals(cons(4, 9),
                              pogo.pogo(c_listMap(c_square, cons(2, 3),
                                                  pogo.land,
                                                  allow_improper_lists=True)))
            self.assertRaises(SchemeError,
                              pogo.pogo,
                              pogo.bounce(c_listMap, c_square,
                                          cons(2, 3), pogo.land))

            self.assertEquals(list(1, 4, 9, 16),
                              pogo.pogo(c_listMap(c_square,
                                                  list(1, 2, 3, 4),
                                                  pogo.land)))


    if __name__ == '__main__':
        unittest.main()
except ImportError:
    pass



try:
	try:
		import pkg_resources
		pkg_resources.declare_namespace(__name__)
	except ImportError:
		import pkgutil
		__path__ = pkgutil.extend_path(__path__, __name__)
except Exception, e:
	import warnings
	warnings.warn(e, RuntimeWarning)

try:
	import modulefinder
	for p in __path__:
		modulefinder.AddPackagePath(__name__, p)
except Exception, e:
	import warnings
	warnings.warn(e, RuntimeWarning)




import os, os.path
import shutil
import stat
import string

import socket
import httplib
import urllib, urlparse
import time

import gzip

#from cache import Cache

from strptime import strptime

"""
Format for media repositories are the following,

media-new.gz
directory1/mediafile.png
directory1/mediafile.mesh

All media which is the same but in a different format must have the same 
basename. See the above example where the file "mediafile" is avaliable in 
both png and mesh format.

The MEDIA is a compressed text file which describes all the media 
avaliable on the server and some metadata about it.
It has the following format
<media file> <last modtime YYYYMMddThhmm> <filesize bytes> <checksumtype>-<checksum>
directory1/mediafile.png 	20060614T0923 1789345 md5-318424ccbd97c644d6baa594284fefe3
directory1/mediafile.mesh 	20060614T0923 7623748 md5-1505b14a78c9e5858edd0f044c4e0062

URLs of media locations are escaped locally using the following system
re.sub('[^a-zA-Z0-9_]', '-', url)
This means that the url "http://media.thousandparsec.net:80/client"
would be escape as      "http---media.thousandparsec.net-80-client"
The URL should always be in full format including the port.

When media is downloaded locally (or distributed via packages) a meta file
is created for each file. This file allows the client to check if the
graphic is up to date. The meta file contains,
<last modtime YYYYMMddThhmm> <filesize bytes> <checksumtype>-<checksum>
20060614T0923 1789345 md5-318424ccbd97c644d6baa594284fefe3

Media can also be distributed via packages. This should be installed in
a shared location such as /usr/share/games/tp, the MEDIA should be 
distributed with the data where the data is split may have to be merged
together (say if the 3d and 2d data from a URL is distributed seperately).

/usr/share/games/tp/http---media.thousandparsec.net-80-client/MEDIA
/usr/share/games/tp/http---media.thousandparsec.net-80-client/mediafile.png
/usr/share/games/tp/http---media.thousandparsec.net-80-client/directory1/mediafile.png.meta
/usr/share/games/tp/ftp---someotherplace.net-21-pub/media-new.gz
...

When a new media is avaliable it will be downloaded to the local 
users home directory.

~/.tp/media/http---media.thousandparsec.net-80-client/media-new.gz
~/.tp/media/http---media.thousandparsec.net-80-client/directory1/mediafile.png
~/.tp/media/http---media.thousandparsec.net-80-client/directory1/mediafile.png.meta
~/.tp/media/ftp---someotherplace.net-21-pub/media-new.gz

If a media file is found in the distribution search locations that is the 
same or newer then the users version. The users version should be removed.
"""

# FIXME: This should be different for different types of file systems.
# FIXME: This should 
searchdirs = [
	os.path.join("usr", "share", "tp"),
	os.path.join("usr", "share", "games", "tp"),
	os.path.join("usr", "local", "share", "tp"),
	os.path.join("usr", "local", "share", "games", "tp"),
	os.path.join("Program Files", "Thousand Parsec", "Media"),
	os.path.join("media"),
	# FIXME: This should only be searched in the development version
	os.path.join("..", "media"),
	"/home/tim/oss/tp/media",
]

import re
def filesafe(url):
	""" Make a URL safe for the filesystem """
	return re.sub('[^a-zA-Z0-9_]', '-', url)

def totime(s):
	return "%02.0f%02.0f%02.0fT%02.0f%02.0f" % strptime(s, "%a, %d %b %Y %H:%M:%S %Z")[0:5]

MEDIA="media-new.gz"

class URLOpener(urllib.FancyURLopener):
	def http_error_default(self, file, socket, code, reason, message):
		raise IOError(code, reason)

class CallbackLimiter(object):
	def __init__(self, realcallback):
		self.realcallback = realcallback
		self.last = 0


	def __call__(self, i, chunksize, maxsize):
		if self.realcallback is None:
			return

		downloaded = chunksize*i
		if not (i == 0 or downloaded == maxsize):
			# Skip if we have had a call back in the last 5 seconds
			if self.last + 5 > time.time():
				return

		self.realcallback(i, chunksize, maxsize)
		self.last  = time.time()

from threadcheck import thread_checker, thread_safe
class Media(object):
	__metaclass__ = thread_checker

	def configdir():
		dirs = [("APPDATA", "Thousand Parsec"), ("HOME", ".tp"), (".", "var")]
		for base, extra in dirs:
			if base in os.environ:
				base = os.environ[base]
				break
			elif base != ".":
				continue

		return os.path.join(base, extra)
	configdir = staticmethod(configdir)

	def hash(self, s):
		"""\
		Get the function needed to do a hash
		"""
		alog, hash = split(s, '-')
		if alog == 'md5':
			import md5
			checksum = (md5.new, hash)
		elif alog == 'sha1':
			import sha
			checksum = (sha.new, hash)
		else:
			raise IOError("Unknown hash algorithm.")
		return hash

	@thread_safe
	def metainfo(self, file):
		"""\
		Return the info in the metafile.
		"""
		modtime, size, checksum = open(file+'.meta', 'r').read().strip().split(' ')
		return modtime, long(size), checksum

	@thread_safe
	def locate(self, file):
		"""\
		Locates a file with a given filename on the filesystem.
		"""
		# Search through the local filesystem and see if we can find the file
		foundhere = []
		for location in self.locations:
			possible = os.path.join(location, file)
			if os.path.exists(possible):
				foundhere.append(possible)
		return foundhere

	@thread_safe
	def newest(self, file):
		"""\
		Returns the newest version of a file.
		"""
		location = None
		curtime = 0
		for possible in self.locate(file):
			try:
				modtime, size, checksum = self.metainfo(possible)
				if modtime > curtime:
					location = possible
					curtime  = modtime
			except (IOError, OSError), e:
				print e
		return location

	def remotetime(self, file):
		"""\
		Gets the remote time of a file.

		Needed to find out if we need to update media-new.gz
		"""
		self.connection.request("HEAD", self.url + file)

		headers = {}
		headers['last-modified'] = self.connection.getresponse().getheader('last-modified')
		return totime(headers['last-modified'])

	def media(self):
		"""\
		"""
		global MEDIA

		# Use the cached version if it's avaliable
		if hasattr(self, '_media'):
			return self._media

		file = self.newest(MEDIA)
		if file is None or not self.connection is None:
			if not file is None:
				modtime, size, checksum = self.metainfo(file)
				# Check if there is a remote version which is new...
				remotetime = self.remotetime(MEDIA)
				if remotetime > modtime:
					# Need to get a new version
					file = self.get(MEDIA)
			else:
				# Need to get a version
				file = self.get(MEDIA)

		media = {}
		for line in gzip.open(file).readlines():
			file, timestamp, size, checksum = line.strip().split()

			media[file] = (timestamp, int(size), checksum)

		# Cache the data so we don't need to reload all the time
		self._media = media

		return media

	def local(self, media, newtime):
		"""\
		Finds a file which is local which has a modtime greater then the give time.
		"""
		foundhere = self.locate(media)

		while len(foundhere) > 0:
			possible = foundhere.pop(0)

			try:
				modtime, size, checksum = self.metainfo(possible)
				if modtime < newtime:
					continue
			except IOError:
				pass

			return possible
		return False

	def give(self, media):
		"""\
		Gets a file to be used. Will use a local version or will download one.
		"""
		mediagz = self.media()
		if not mediagz.has_key(media):
			raise IOError("No such media file exists!")

		modtime = mediagz[media][0]
		# See if we have a local version we can use
		location = self.local(media, modtime)
		if not location is False:
			return location	
		
		# Looks like we will have to download a peice of media
		return self.get(media)

	def get(self, file, callback=None):
		"""\
		Gets a file from the remote server.
		"""
		global MEDIA

		# Where the file will be downloaded too
		local_location  = os.path.join(self.locations[-1], file)
		# Where the file is on the remote server
		remote_location = urlparse.urljoin(self.url, file)

		# If the file already exists we better remove it
		if os.path.exists(local_location):
			os.unlink(local_location)

		dir = os.path.dirname(local_location)
		if not os.path.exists(dir):
			os.makedirs(dir)

		if 1:		
#		try:
			# Download the file
			(trash, message) = self.getter.retrieve(remote_location, local_location, CallbackLimiter(callback))

			if file != MEDIA:
				mediagz = self.media()
				# Check the checksum

				# Check the filesize

				# Create the metafile from the mediagz info
				open(local_location+".meta", 'w').write("%s %s %s" % mediagz[file])
			else:
				# Have to generate our own data
				open(local_location + ".meta", 'w').write("%s %s %s" % (totime(message.getheader('last-modified')), 0, 'None'))
			return local_location
#		except IOError, e:
#			print e
#			return False

	def __init__(self, url, configdir=None):
		"""\
		Everything must be there, even if the port is the default.

		serverurl is the URL of the server to download the media from

		configdir is where the cache will be stored.
		mediatypes is the type of media which the client needs.
		"""
		self.url = url
		# Make the URL safe for filesystems
		safeurl = filesafe(url)

		# FIXME: Check the serverurl is valid, IE It's in the full form.
		if configdir == None:
			configdir = self.configdir()
		userdir = os.path.join(configdir, "media", safeurl)

#		# Make sure the user dir exists
#		if os.path.exists(userdir) and new:
#			shutil.rmtree(userdir)
		if not os.path.exists(userdir):
			os.makedirs(userdir)

		# Find all the locations which have media relavent to this URL.
		self.locations = []
		global searchdirs
		for location in searchdirs:
			full = os.path.join(location, safeurl)
			if os.path.exists(full):
				self.locations.append(full)

		# Append the users "localdir"
		self.locations.append(userdir)

	@property
	def connection(self):
		if not hasattr(self, '_connection'):
			type, host, self.basepath, t, t, t = urlparse.urlparse(self.url)
			self._connection = getattr(httplib, "%sConnection" % type.upper())(host)
		return self._connection

	@property
	def getter(self):
		if not hasattr(self, '_getter'):
			self._getter = URLOpener()
		return self._getter

	def getpossible(self, media_types):
		"""
		Gets the Media description file from the http server.
		"""
		for file in self.media():
			for possible in media_types:
				if file.endswith(possible):
					yield file

if __name__ == "__main__":
	import sys
	media_cache = Media(sys.argv[1])

	files = {}
	for file in media_cache.getpossible(['png', 'gif', 'jpg']):
		print file
		media_cache.give(file)
